@@ -56,6 +56,8 @@
 #' 
 #' @aliases print.calibrate
 #' 
+#' @importFrom stats coef formula lm lm.fit model.extract model.frame 
+#' @importFrom stats model.matrix model.response na.fail qf qt resid terms var
 #' @export
 #'
 #' @note The function \code{invest} is more general, but based on numerical
@@ -99,6 +101,7 @@ calibrate <- function(object, ...) {
   UseMethod("calibrate")
 }
 
+
 #' @rdname calibrate
 #' @export
 #' @method calibrate default
@@ -124,19 +127,22 @@ calibrate.default <- function(object, y0,
                  "' not of same length.", sep = ""), call. = FALSE)
     }
   } else {
-    stop(paste(deparse(substitute(object)), 
-               "' is not a valid matrix, list, or data frame.", sep = ""),
+    stop("'", paste(deparse(substitute(object)), 
+                    "' is not a valid matrix, list, or data frame.", sep = ""),
          call. = FALSE)
   }
   eta <- mean(y0)  # mean of new observations
   m <- length(y0)  # number of new observations
   if (mean.response && m > 1) stop("Only one mean response value allowed.")
-  b <- unname(coef(z <- lm(y ~ x, data = data.frame(x, y))))  # coefficients
-  n <- length(e <- resid(z))  # sample size and residuals
-  df <- (df1 <- n - 2) + (df2 <- m - 1)  # degrees of freedom
-  var1 <- Sigma(z)^2  # stage 1 variance estimate
+  
+  # Fit a simple linear regression model and compute necessary components
+  z <- lm.fit(cbind(1, x), y)
+  b <- unname(z$coefficients)
+  n <- length(r <- z$residuals)  # sample size and residuals
+  DF <- (DF1 <- n - 2) + (DF2 <- m - 1)  # degrees of freedom
+  var1 <- sum(r ^ 2) / z$df.residual  # stage 1 variance estimate
   var2 <- if (m == 1) 0 else var(y0)  # stage 2 variance estimate
-  var.pooled <- (df1 * var1 + df2 * var2) / df  # pooled estimate of variance
+  var.pooled <- (DF1 * var1 + DF2 * var2) / DF  # pooled estimate of variance
   sigma.pooled <- sqrt(var.pooled)  # sqrt of pooled variance estimate
   ssx <- sum((x - mean(x))^2)  # sum-of-squares for x, Sxx
   x0.mle <- (eta - b[1L])/b[2L]  # MLE of x0
@@ -183,7 +189,9 @@ calibrate.default <- function(object, y0,
       }
       
     }
-    res <- list("estimate" = x0.mle, "lower" = lwr, "upper" = upr, 
+    res <- list("estimate" = x0.mle, 
+                "lower"    = lwr, 
+                "upper"    = upr, 
                 "interval" = interval)
   
   }
@@ -199,8 +207,11 @@ calibrate.default <- function(object, y0,
     }
 
     # Store results in a list
-    res <- list("estimate" = x0.mle, "lower" = x0.mle - crit * se,
-                "upper" = x0.mle + crit * se, "se" = se, "interval" = interval)
+    res <- list("estimate" = x0.mle, 
+                "lower"    = x0.mle - crit * se,
+                "upper"    = x0.mle + crit * se, 
+                "se"       = se, 
+                "interval" = interval)
     
   } 
   
@@ -210,6 +221,7 @@ calibrate.default <- function(object, y0,
   
 } 
 
+
 #' @rdname calibrate
 #' @export
 #' @method calibrate formula
@@ -231,9 +243,133 @@ calibrate.formula <- function(formula, data = NULL, ..., subset,
   calibrate(cbind(x, y), ...)
 } 
 
+
+#' #' @rdname calibrate
+#' #' @export
+#' #' @method calibrate lm
+#' calibrate.lm <- function(object, ...) {
+#'   calibrate(formula(object), data = eval(object$call$data), ...)
+#' } 
+
+
 #' @rdname calibrate
 #' @export
 #' @method calibrate lm
-calibrate.lm <- function(object, ...) {
-  calibrate(formula(object), data = eval(object$call$data), ...)
-} 
\ No newline at end of file
+calibrate.lm <- function(object, y0, interval = c("inversion", "Wald", "none"), 
+                         level = 0.95, mean.response = FALSE, 
+                         adjust = c("none", "Bonferroni", "Scheffe"), k, ...) {
+  
+  # Check model formula for correctness
+  xname <- all.vars(formula(object)[[3L]])
+  yname <- all.vars(formula(object)[[2L]])
+  if (length(xname) != 1L) {
+    stop("Only one independent variable allowed.")
+  }
+  if (length(yname) != 1L) {
+    stop("Only one dependent variable allowed.")
+  }
+  
+  # Check for intercept using terms object from model fit. Alternatively, this
+  # can also be checked by testing if the first column name in model.matrix is 
+  # equal to "(Intercept)".
+  if (!attr(terms(object), "intercept")) {
+    stop(paste(deparse(substitute(object)), "must contain an intercept."))
+  }
+  
+  # Extract x values and y values from model frame
+  mf <- model.frame(object)
+  if (ncol(mf) != 2) {
+    stop("calibrate only works for the simple linear regression model.")
+  } 
+  x <- model.matrix(object)[, 2]
+  y <- model.response(mf)
+
+  # Eta - mean response or mean of observed respone values
+  eta <- mean(y0)  # mean of new observations
+  m <- length(y0)  # number of new observations
+  if (mean.response && m > 1) stop("Only one mean response value allowed.")
+  
+  # Fit a simple linear regression model and compute necessary components
+  b <- unname(object$coefficients)
+  n <- length(r <- object$residuals)  # sample size and residuals
+  DF <- (DF1 <- n - 2) + (DF2 <- m - 1)  # degrees of freedom
+  var1 <- sum(r ^ 2) / object$df.residual  # stage 1 variance estimate
+  var2 <- if (m == 1) 0 else var(y0)  # stage 2 variance estimate
+  var.pooled <- (DF1 * var1 + DF2 * var2) / DF  # pooled estimate of variance
+  sigma.pooled <- sqrt(var.pooled)  # sqrt of pooled variance estimate
+  ssx <- sum((x - mean(x))^2)  # sum-of-squares for x, Sxx
+  x0.mle <- (eta - b[1L])/b[2L]  # MLE of x0
+  
+  # Return point estimate only
+  interval <- match.arg(interval)
+  if (interval == "none") return(x0.mle)
+  
+  # Adjustment for simultaneous intervals
+  adjust <- match.arg(adjust)  # FIXME: Does simultaneous work for m > 1?
+  crit <- if (m != 1 || adjust == "none") qt((1 + level)/2, n+m-3) else {
+    switch(adjust,
+           "Bonferroni" = qt((level + 2*k - 1) / (2*k), n+m-3),
+           "Scheffe"    = sqrt(k * qf(level, k, n+m-3)))
+  }
+  
+  # Inversion interval --------------------------------------------------------
+  if (interval == "inversion") { 
+    
+    c1 <- b[2L]^2 - (sigma.pooled^2 * crit^2)/ssx
+    c2 <- if (mean.response) {
+      c1/n + (eta - mean(y))^2/ssx
+    } else {
+      c1*(1/m + 1/n) + (eta - mean(y))^2/ssx
+    }
+    c3 <- b[2L] * (eta - mean(y))
+    c4 <- crit * sigma.pooled
+    
+    # FIXME: catch errors and throw an appropriate warning
+    if (c1 < 0 && c2 <= 0) {
+      
+      warning("The calibration line is not well determined.", call. = FALSE)
+      lwr <- -Inf
+      upr <- Inf
+      
+    } else {
+      
+      lwr <- mean(x) + (c3 - c4*sqrt(c2))/c1
+      upr <- mean(x) + (c3 + c4*sqrt(c2))/c1
+      if (c1 < 0 && c2 > 0) {
+        stop(paste("The calibration line is not well determined. The resulting \nconfidence region is the union of two semi-infinite intervals:\n(", -Inf, ",", 
+                   round(upr, 4), ") U (", round(lwr, 4), ",", Inf, ")"), 
+             call. = FALSE)
+      }
+      
+    }
+    res <- list("estimate" = x0.mle, 
+                "lower"    = lwr, 
+                "upper"    = upr, 
+                "interval" = interval)
+    
+  }
+  
+  # Wald interval  
+  if (interval == "Wald") { 
+    
+    # Compute standard error for Wald interval
+    se <- if (mean.response) {
+      abs((sigma.pooled/b[2]))*sqrt((1/n + (x0.mle - mean(x))^2/ssx))
+    } else {
+      abs((sigma.pooled/b[2]))*sqrt((1/m + 1/n + (x0.mle - mean(x))^2/ssx))
+    }
+    
+    # Store results in a list
+    res <- list("estimate" = x0.mle, 
+                "lower"    = x0.mle - crit * se,
+                "upper"    = x0.mle + crit * se, 
+                "se"       = se, 
+                "interval" = interval)
+    
+  } 
+  
+  # Assign class label and return results
+  class(res) <- "invest"
+  return(res)
+  
+} 
